local addon, shared = ...

-- UI Setup --

--Create context for rendering
local context = UI.CreateContext("Context")

local slideDownFrame = UI.CreateFrame("Frame", "slideDownFrame", context)

slideDownFrame:EventAttach(Event.UI.Input.Mouse.Left.Down, function(self, h)
	if SlideDown_Settings.locked == false then
		self.MouseDown = true
		local mouseData = Inspect.Mouse()
		self.sx = mouseData.x - slideDownFrame:GetLeft() - slideDownFrame:GetWidth()/2
		self.sy = mouseData.y - slideDownFrame:GetTop()
	end
end, "Event.UI.Input.Mouse.Left.Down")

slideDownFrame:EventAttach(Event.UI.Input.Mouse.Cursor.Move, function(self, h)
	if self.MouseDown then
		local nx, ny
		local mouseData = Inspect.Mouse()
		nx = mouseData.x - self.sx
		ny = mouseData.y - self.sy
		slideDownFrame:SetPoint("TOPCENTER", UIParent, "TOPLEFT", nx,ny)
	end
end, "Event.UI.Input.Mouse.Cursor.Move")

slideDownFrame:EventAttach(Event.UI.Input.Mouse.Left.Up, function(self, h)
	if self.MouseDown then
		self.MouseDown = false
		SlideDown_Settings.xPos = (slideDownFrame:GetLeft() + slideDownFrame:GetWidth()/2)/UIParent:GetWidth()
		SlideDown_Settings.yPos = slideDownFrame:GetTop()/UIParent:GetHeight()
	end
end, "Event.UI.Input.Mouse.Left.Up")

local time_5m_marker = UI.CreateFrame("Text", "time_5m_marker", slideDownFrame)
local time_3m_marker = UI.CreateFrame("Text", "time_3m_marker", slideDownFrame)
local time_1m_marker = UI.CreateFrame("Text", "time_1m_marker", slideDownFrame)
local time_30s_marker = UI.CreateFrame("Text", "time_30s_marker", slideDownFrame)
local time_15s_marker = UI.CreateFrame("Text", "time_15s_marker", slideDownFrame)
local time_5s_marker = UI.CreateFrame("Text", "time_5s_marker", slideDownFrame)
local time_1s_marker = UI.CreateFrame("Text", "time_1s_marker", slideDownFrame)

local default_settings = {
	travelLength =		450,		--Travel length is first width.
	iconHeight = 		30,			--Icon height is first height.
	movementFlipped =	false,		--Right-to-Left or Top-to-Bottom by default. If true, changes movement direction.
	horizontal = true,				--Horizontal mode by default. If false, changes to vertical mode.
	bgRed = 0.0,
	bgGreen = 0.8,
	bgBlue = 1.0,
	xPos = 0.8,
	yPos = 0.5,
	logarithmic = true,
	locked = false,
	trans = true
}

local label_marker = { 
	["5m"] = time_5m_marker,
	["3m"] = time_3m_marker,
	["1m"] = time_1m_marker,
	["30s"] = time_30s_marker,
	["15s"] = time_15s_marker,
	["5s"] = time_5s_marker,
	["1s"] = time_1s_marker
}

local time_marker = { 
	[300] = time_5m_marker,
	[180] = time_3m_marker,
	[60] = time_1m_marker,
	[30] = time_30s_marker,
	[15] = time_15s_marker,
	[5] = time_5s_marker,
	[1] = time_1s_marker
}
	
--Max cooldown time
local maxCooldownTime = 600

--Store actual things on cooldown. Key = Ability ID. Value = Slider to show on bar.
local activeCooldowns = {}
local activeCooldownList = {}

local function calcBarPoint(cooldownRemaining)
	if SlideDown_Settings.logarithmic then
		barPoint = (math.log(((cooldownRemaining / maxCooldownTime) * 1999) + 1) / math.log(2000))
	else
		barPoint = cooldownRemaining / maxCooldownTime
	end
	return barPoint
end
-- Overall UI Functions --

--Refreshes the UI in respect to the variables users can change.
--Runs when options are used to change the UI.
local function refreshUI()
	slideDownFrame:SetPoint("TOPCENTER", UIParent, SlideDown_Settings.xPos, SlideDown_Settings.yPos)
	
	actualHeight = 0
	actualLength = 0
	if SlideDown_Settings.horizontal then
		actualHeight = SlideDown_Settings.iconHeight
		actualLength = SlideDown_Settings.travelLength
		
		for timeAmount, marker in pairs(time_marker) do
			marker:SetPoint("CENTER", slideDownFrame, calcBarPoint(tonumber(timeAmount)), 0.5)
		end
	else
		actualHeight = SlideDown_Settings.travelLength
		actualLength = SlideDown_Settings.iconHeight
		
		for timeAmount, marker in pairs(time_marker) do
			marker:SetPoint("CENTER", slideDownFrame, 0.5, calcBarPoint(tonumber(timeAmount)))
		end
	end

	slideDownFrame:SetHeight(actualHeight)
	slideDownFrame:SetWidth(actualLength)
	
	for _, marker in pairs(time_marker) do
		marker:SetHeight(SlideDown_Settings.iconHeight)
		marker:SetWidth(SlideDown_Settings.iconHeight)
		marker:SetFontSize(SlideDown_Settings.iconHeight * 0.5)
	end
	
	slideDownFrame.bg:SetBackgroundColor(SlideDown_Settings.bgRed, SlideDown_Settings.bgGreen, SlideDown_Settings.bgBlue)
end


--Initial UI setup. Runs after saved variables are loaded.
local function setUpUI()
	--Set up the SlideDown main frame and background.
	slideDownFrame.bg = UI.CreateFrame("Frame", "slideDownFrame.bg", slideDownFrame)
	slideDownFrame.bg:SetPoint("TOPLEFT", slideDownFrame, "TOPLEFT")
	slideDownFrame.bg:SetPoint("BOTTOMRIGHT", slideDownFrame, "BOTTOMRIGHT")
	slideDownFrame.bg:SetLayer(-1)
	
	for label, marker in pairs(label_marker) do
		marker:SetText(label)
	end
	
	refreshUI()
	
	for _, marker in pairs(time_marker) do
		marker:SetVisible(true)
	end
	slideDownFrame.bg:SetVisible(true)
end

-- Saved Variables Handlers --

--Loads saved variables for different settings, if they exist.
--Then set up UI after loading.
local function loadVariables(h,a)
	if a == "SlideDown" then
		if SlideDown_Settings == nil then SlideDown_Settings = {} end
		
		for k,v in pairs(default_settings) do
			if SlideDown_Settings[k] == nil then
				SlideDown_Settings[k] = v
			end
		end
	
		setUpUI()
		slideDownFrame.bg:SetVisible(not SlideDown_Settings.trans)
	end
end


local ABILITY_IGNORE = {}

-- Event Handlers --

--If the cooldown is real (not global cooldown), then make it a slider and store it in the active buff table.
--Runs for each cooldown in each Event.Ability.Cooldown.Begin.
local function cooldownBegin(cooldownList)
	local abilityDtl = Inspect.Ability.New.Detail(cooldownList)
	for k,v in pairs(abilityDtl) do
		local slider
		--Check to see if real cooldown or GCD
		if v.cooldown == nil then
			ABILITY_IGNORE[k] = true
		else
			if (v.currentCooldownDuration <= v.cooldown + 0.0001) and
				(v.currentCooldownDuration >= v.cooldown - 0.0001) then
				if activeCooldowns[k] == nil then
					slider = UI.CreateFrame("Texture", "slider", slideDownFrame)
					slider:SetHeight(SlideDown_Settings.iconHeight)
					slider:SetWidth(SlideDown_Settings.iconHeight)
					activeCooldowns[k] = slider
				else
					slider = activeCooldowns[k]
				end
				slider:SetVisible(true)
				slider.active = true
				activeCooldownList[k] = true
				
				--The icons that slide along the SlideDown bar are "sliders".
				--Currently, Sliders are squares as tall as the bar.
				
				slider:SetPoint("CENTER", slideDownFrame, "CENTERRIGHT")
				slider:SetTexture("Rift", v.icon)
			end
		end
	end
end

-- Refresh Functions --

--For each active cooldown, update it's slider.
--Runs every 1/10s

local lastUpdate = 0

local function slideDown(h)
	local curTime = Inspect.Time.Real()
	local cooldownRemaining
	local sliderXPos = 0
	local sliderYPos = 0
	local barPoint
	
	if curTime - lastUpdate > 0.1 then
		local abilityDetail = Inspect.Ability.New.Detail(activeCooldownList)
		for k,v in pairs(abilityDetail) do
			if activeCooldowns[k].active then
				if v.currentCooldownRemaining then --check. Sometimes is null for last loop.
				
					cooldownRemaining = v.currentCooldownRemaining
					if (cooldownRemaining > maxCooldownTime) then
						cooldownRemaining = maxCooldownTime
					end
					
					--print(detail.name)
					--barPoint defines the function which models the movement of the sliders.
					barPoint = calcBarPoint(cooldownRemaining)
					--print(barPoint)
					
					if SlideDown_Settings.movementFlipped then
						barPoint = 1.0 - barPoint
					end
										
					if SlideDown_Settings.horizontal then
						sliderXPos = barPoint
						sliderYPos = 0.5
					else
						sliderXPos = 0.5
						sliderYPos = barPoint
					end
					
					activeCooldowns[k]:SetPoint("CENTER", slideDownFrame, sliderXPos, sliderYPos)
				else	
					activeCooldowns[k]:SetVisible(false)
					activeCooldowns[k].active = false
					activeCooldownList[k] = nil
				end
			end
		end
		lastUpdate = curTime
	end
end



-- Slash Command Handlers --

--Distribute this request to it's proper function.
--Runs when a /slidedown command is entered
local function slashinput(h,input)
	local args = {}
	for token in string.gmatch(input, "[^%s]+") do
		table.insert(args, token)
	end

	if args[1] == "pos" then
		tempXPos = tonumber(args[2])
		tempYPos = tonumber(args[3])
		if tempXPos and tempYPos then
			--print(xPos)
			--print(yPos)
			SlideDown_Settings.xPos = tempXPos
			SlideDown_Settings.yPos = tempYPos
			refreshUI()
		end
	elseif args[1] == "horiz" then
		if args[2] == "false" then
			SlideDown_Settings.horizontal = false
		else
			SlideDown_Settings.horizontal = true
		end
		refreshUI()
	elseif args[1] == "flipped" then
		if args[2] == "false" then
			SlideDown_Settings.movementFlipped = false
		else
			SlideDown_Settings.movementFlipped = true
		end
	elseif args[1] == "logarithmic" then
		if args[2] == "false" then
			SlideDown_Settings.logarithmic = false
		else
			SlideDown_Settings.logarithmic = true
		end
	elseif args[1] == "locked" then
		if args[2] == "false" then
			SlideDown_Settings.locked = false
		else
			SlideDown_Settings.locked = true
		end
	elseif args[1] == "size" then
		tempTrav = tonumber(args[2])
		tempIcon = tonumber(args[3])
		if tempTrav and tempIcon then
			SlideDown_Settings.travelLength = tempTrav
			SlideDown_Settings.iconHeight = tempIcon
			refreshUI()
		end
	elseif args[1] == "color" then
		red = tonumber(args[2])
		green = tonumber(args[3])
		blue = tonumber(args[4])
		if red and green and blue then
			if (red <= 1.0 and red >= 0.0) and 
			(green <= 1.0 and green >= 0.0) and 
			(blue <= 1.0 and blue >= 0.0) then
				SlideDown_Settings.bgRed = red
				SlideDown_Settings.bgGreen = green
				SlideDown_Settings.bgBlue = blue
				refreshUI()
			end
		end
	elseif args[1] == "reset" then
		SlideDown_Settings = {}
		for k,v in pairs(default_settings) do
			if SlideDown_Settings[k] == nil then
				SlideDown_Settings[k] = v
			end
		end

		refreshUI()
elseif args[1] == "trans" then
		if args[2] == "false" then
			SlideDown_Settings.trans = false
		else
			SlideDown_Settings.trans = true
		end
		slideDownFrame.bg:SetVisible(not SlideDown_Settings.trans)
	end
end

local function Event_Unit_Availability_Full(h,t)
	for k,v in pairs(t) do
		if v == "player" then
			local cdList = {}
			Command.Event.Detach(Event.Unit.Availability.Full, nil, nil, nil, addon.identifier)
			for k,v in pairs(Inspect.Ability.New.Detail(Inspect.Ability.New.List())) do
				if v.cooldown and v.currentCooldownRemaining then
					cdList[k] = true
				end
			end
			cooldownBegin(cdList)
		end
	end
end


-- Event Hooks --

Command.Event.Attach(Event.Unit.Availability.Full, Event_Unit_Availability_Full, "Event.Unit.Availability.Full")
Command.Event.Attach(Event.Addon.SavedVariables.Load.End, loadVariables, "Load saved variables")

--Every frame, update the positions of active sliders.
Command.Event.Attach(Event.System.Update.Begin, slideDown, "slideDown")

--The cooldown begin event passes a table of all abilities that went on cooldown.
--key = ability ID, value = cooldown time remaining.
Command.Event.Attach(Event.Ability.New.Cooldown.Begin,
		function (h,params)
			for abilityID, cooldownTime in pairs(params) do
				if ABILITY_IGNORE[abilityID] then
					params[abilityID] = nil
				end
			end
			cooldownBegin(params)
		end, 
		"cooldownBegin"
)

--The cooldown begin event passes a table of all abilities that came off cooldown.
--key = ability ID, value = cooldown time remaining (0, heh).
Command.Event.Attach(Event.Ability.New.Cooldown.End, 
		function (h,params)
			for abilityID, cooldownTime in pairs(params) do
				if activeCooldowns[abilityID] then
					activeCooldowns[abilityID]:SetVisible(false)
					activeCooldowns[abilityID].active = false
					activeCooldownList[abilityID] = nil
				end
			end
		end, 
		"cooldownEnd"
)

-- Slash Command Register --

Command.Event.Attach(Command.Slash.Register("slidedown"), slashinput, "Slash command providing access to SlideDown options.")